#!/usr/bin/env python3.2
## Copyright © LIMSI-CNRS (2013)
##
## contributor(s) : Jorge Gascon, Damien Touraine, David Poirier-Quinot,
## Laurent Pointal, Julian Adenauer, 
## 
## This software is a computer program whose purpose is to distribute
## blender to render on CAVE(TM) device systems.
## 
## This software is governed by the CeCILL  license under French law and
## abiding by the rules of distribution of free software.  You can  use, 
## modify and/ or redistribute the software under the terms of the CeCILL
## license as circulated by CEA, CNRS and INRIA at the following URL
## "http://www.cecill.info". 
## 
## As a counterpart to the access to the source code and  rights to copy,
## modify and redistribute granted by the license, users are provided only
## with a limited warranty  and the software's author,  the holder of the
## economic rights,  and the successive licensors  have only  limited
## liability. 
## 
## In this respect, the user's attention is drawn to the risks associated
## with loading,  using,  modifying and/or developing or reproducing the
## software by the user in light of its specific status of free software,
## that may mean  that it is complicated to manipulate,  and  that  also
## therefore means  that it is reserved for developers  and  experienced
## professionals having in-depth computer knowledge. Users are therefore
## encouraged to load and test the software's suitability as regards their
## requirements in conditions enabling the security of their systems and/or 
## data to be ensured and,  more generally, to use and operate it in the 
## same conditions as regards security. 
## 
## The fact that you are presently reading this means that you have had
## knowledge of the CeCILL license and that you accept its terms.
## 

import os
import sys
import time
import socket
import subprocess
import getpass
import pickle

class BlenderCave_Configuration:
    def __init__(self):
        if sys.platform.startswith("win"):
            self._path = os.path.join(os.environ['APPDATA'])
        elif sys.platform == 'darwin':
            self._path = os.path.join(os.path.expanduser('~'), 'Library', 'Application Support')
        else: # Unix ...
            self._path = os.path.join(os.path.expanduser('~'), '.config')
        self._path = os.path.join(self._path, 'blender')

        if not os.path.isdir(self._path):
            os.makedirs(self._path)

        self._file = os.path.join(self._path, 'Blender CAVE.conf')
        self.load()

    def load(self):
        try:
            with open(self._file, 'rb') as file:
                self._information = pickle.load(file)
        except:
            self._information = {'blend_file'     : None,
                                 'config'         : None,
                                 'config_file'    : None,
                                 'log_run'        : False,
                                 'previous_state' : False}

    def get(self):
        return self._information

    def save(self, arguments, state, log_run):
        self._information = {}
        for element in ['blend_file', 'config', 'config_file']:
            if hasattr(arguments, element):
                self._information[element] = getattr(arguments, element)
        self._information['previous_state'] = state
        self._information['log_run'] = log_run
        try:
            with open(self._file, 'wb') as file:
                pickle.dump(self._information, file)
        except IOError:
            pass

main_config = BlenderCave_Configuration()
config_file_informations = main_config.get()

import argparse

class _Parser(argparse.ArgumentParser):

    def __init__(self):
        super(_Parser, self).__init__("usage: %prog")
        self._errors = []

    def error(self, msg):
        self._errors.append(msg)
        pass

    def getErrors(self):
        return self._errors

parser = _Parser()
parser.add_argument("--version",      dest="version",      help="Log blenderplayer running",     action="store_true", default=False)
parser.add_argument("-f", "--file",   dest="blend_file",   help="Blender file to load",          default=config_file_informations['blend_file'])
parser.add_argument("-c", "--config", dest="config",       help="Current configuration to run",  default=config_file_informations['config'])
parser.add_argument("--config-file",  dest="config_file",  help="Configuration File",            metavar="FILE",      default=config_file_informations['config_file'])
parser.add_argument("-l",             dest="log_run",      help="Toggle log  running",           action="store_true", default=False)
parser.add_argument("-k",             dest="kill",         help="Kill the simulation",           action="store_true", default=False)
parser.add_argument("-s",             dest="start",        help="Start the simulation",          action="store_true", default=False)
parser.add_argument("-d",             dest="display_conf", help="Display current configuration", action="store_true", default=False)

previous_state = config_file_informations['previous_state']
log_run        = config_file_informations['log_run']
del(config_file_informations)

environments = parser.parse_args(sys.argv[1:])

if environments.display_conf:
    info = main_config.get()
    print('Blender file: ', info['blend_file'])
    print('Configuration file:', info['config_file']) 
    print('Current configuration:', info['config'])
    print('Login of blender run: ', info['log_run'])
    print('Previous state: ', info['previous_state'])
    sys.exit()

def prepareFile(file):
    return os.path.realpath(os.path.expanduser(file))

if environments.version:
    exec(open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'blender_cave', 'version.py')).read())
    print('BlenderCave version: ', version)
    sys.exit()

if len(parser.getErrors()) > 0:
    print(sys.argv[0], ' wrong arguments:')
    for message in parser.getErrors():
        print(message)
    sys.exit()

if environments.config is None:
    print(sys.argv[0], ': missing config')
    sys.exit()

if environments.config_file is None:
    print(sys.argv[0], ': missing configuration file')
    sys.exit()

environments.config_file = prepareFile(environments.config_file)

if environments.blend_file is None:
    print(sys.argv[0], ': missing blender file')
    sys.exit()

if not os.path.isfile(environments.blend_file):
    print('invalid blender file : ', environments.blend_file)
    sys.exit()

environments.blend_file = prepareFile(environments.blend_file)

if environments.kill:
    state=False
elif environments.start:
    state=True
elif previous_state:
    state = False
else:
    state = True

if environments.log_run:
    if log_run:
        log_run = False
    else:
        log_run = True

main_config.save(environments, state, log_run)

import xml.sax.handler

class Parser(xml.sax.handler.ContentHandler):
    def __init__(self, file_name):
        self._parser = xml.sax.make_parser()
        self._parser.setContentHandler(self)

        self._inside_blender_cave = False
        self._inside_starter      = False
        self._inside_system       = False
        self._inside_config_path  = False

        self._system_paths  = []
        self._starter_paths = []
        self._system_root   = None
        self._starter_root  = None
        self._other_level   = -1

        try:
            self._parser.parse(file_name)
        except ValueError:
            pass
        except SystemExit:
            sys.exit()
        except: # Don't try to fully parse the file !
            import traceback
            traceback.print_exc()
            sys.exit()
        if self._starter_root is not None:
            self._root = self._starter_root
        elif self._system_root is not None:
            self._root = self._system_root
        else:
            self._root = os.path.dirname(os.path.realpath(__file__))
        self._root = self._evaluateInput(self._root)
        if len(self._starter_paths) == 0:
            self._paths = self._system_paths
        else:
            self._paths = self._starter_paths

    def _evaluateInput(self, variable):
        if variable is None:
            self.error('Missing variable !')
        result = ''
        for index, element in enumerate(variable.split('`')):
            if index % 2 == 1:
                evaluation = None
                try:
                    evaluation = eval(element)
                except SyntaxError:
                    pass
                except NameError:
                    pass
                if evaluation == None:
                    try:
                        import subprocess
                        proc = subprocess.Popen(element, stdout=subprocess.PIPE, shell=True)
                        (out, err) = proc.communicate()
                        out = out.decode('UTF-8')
                        if out.strip() != '':
                            evaluation = out
                    except:
                        pass
                if evaluation == None:
                    self.error('invalid evaluation of `' + element + '`')
                result += evaluation.strip()
            else:
                result += element
        return result

    def error(self, msg):
        print('line ' + str(self._locator.getLineNumber()) + ' of ' + self._locator.getSystemId() + ': ' + msg)
        sys.exit()

    def startElement(self, name, attrs):
        if name == 'blender_cave':
            if self._inside_blender_cave:
                self.error('Cannot include "blender_cave" section inside "blender_cave" section !')
            self._inside_blender_cave = True
        elif not self._inside_blender_cave:
            self.error('Everything must be include inside a "blender_cave" section !')
        elif self._other_level >= 0:
            self._other_level += 1
        elif name == 'system':
            if self._inside_system or self._inside_starter:
                self.error('Cannot define a "system" section inside "system" or "starter" section !')
            self._inside_system = True
            self._system_paths  = []
            try :
                self._system_root = attrs['root']
            except KeyError:
                self._system_root = None
        elif name == 'starter':
            if self._inside_system or self._inside_starter: 
                self.error('Cannot define a "starter" section inside "system" or "starter" section !')
            self._inside_starter = True
            self._starter_paths  = []
            try :
                self._starter_root = attrs['root']
            except KeyError:
                self._starter_root = None
        elif name == 'config_path':
            if not self._inside_system and not self._inside_starter: 
                self.error('Cannot define a "config_path" section outside "system" or "starter" section !')
            self._inside_config_path = True
        else:
            self._other_level = 0

    def characters(self, string):
        if self._inside_config_path:
            paths = self._evaluateInput(string).split(os.pathsep)
            if self._inside_system:
                self._system_paths += paths
            elif self._inside_starter:
                self._starter_paths += paths

    def endElement(self, name):
        if name == 'blender_cave':
            self._inside_blender_cave = False
        elif self._other_level >= 0:
            self._other_level -= 1
        elif name == 'system':
            self._inside_system = False
        elif name == 'starter':
            self._inside_starter = False
        elif name == 'config_path':
            self._inside_config_path = False

parser = Parser(environments.config_file)

root        = parser._root

config_file = os.path.basename(environments.config_file)
config_path = os.path.dirname(environments.config_file)

if config_path not in parser._paths:
    parser._paths += [config_path]

if len(root) == 0:
    print('Invalid global BlenderCave configuration file: missing root of blender_cave !')
    sys.exit()

if not os.path.isdir(os.path.join(root, 'utils')):
    print('Invalid global BlenderCave configuration file: root must point to directory containing blender_cave !')
    sys.exit()

try:
    import imp
    (file, file_name, data) = imp.find_module('launcher', [os.path.join(root, 'utils')])
except:
    print('Cannot load the launcher module !')
    sys.exit()

module = imp.load_module('launcher', file, file_name, data)
launcher = module.Launcher(parser._paths, config_file)
configuration = launcher.getConfiguration()

starter = configuration['starter']
config  = environments.config

if config not in starter['configs']:
    print("Cannot find config '" + config + "' inside XML configuration file !")
    print("Valid configs : ", ', '.join(starter['configs'].keys()))
    sys.exit()

background = (len(starter['configs'][config]) > 1)

if not background:
    main_config.save(environments, False, log_run)

def stripAnchor(path):
    if path is not None and starter['anchor'] is not None:
        relpath = os.path.relpath(path, starter['anchor'])
        if not relpath.startswith('..'):
            return relpath
    return path

default = {'python'     : sys.executable,
           'root'       : root,
           'config_file': config_file,
           'config_path': config_path,
           'blend_file' : stripAnchor(environments.blend_file)}

screen_configs = []
for screen_name in starter['configs'][config]:
    if screen_name not in configuration['screens']:
        print("Cannot find screen '" + screen_name + "' inside XML configuration file !")
        sys.exit()
    screen    = configuration['screens'][screen_name]
    computer_name = screen['computer']
    if computer_name not in configuration['computers']:
        print("Cannot find computer '" + computer_name + "' inside XML configuration file !")
        sys.exit()
    computer = configuration['computers'][computer_name]
    system   = computer['system']

    screen_config = {'screen'       : screen_name,
                     'config'       : config,
                     'blend_file'   : default['blend_file'],
                     'environments' : []}

    
    for key, value in system['blenderplayer']['environments'].items():
        screen_config['environments'].append(key + '=' + value)

    if 'login' in system:
        login = system['login']
        if ('remote_command' in login) and (login['remote_command'] is not None):
            screen_config['run'] = login['remote_command']
        else:
            screen_config['run'] = ''
        if ('python' in login) and (login['python'] is not None):
            screen_config['python'] = login['python']
        else:
            screen_config['python'] = default['python']
    else:
        screen_config['run'] = ''
        screen_config['python'] = default['python']

    if ('root' in system) and (system['root'] is not None):
        screen_config['root'] = system['root']
    else:
        screen_config['root'] = default['root']

    if ('config_file' in system) and (system['config_file'] is not None):
        screen_config['config_file'] = system['config_file']
    else:
        screen_config['config_file'] = default['config_file']

    if ('config_paths' in system) and (system['config_paths'] is not None):
        screen_config_paths = system['config_paths']
    else:
        screen_config_paths = default['config_path']
    if type(screen_config_paths) is list:
        screen_config['config_path'] = os.pathsep.join(screen_config_paths)
    else:
        screen_config['config_path'] = screen_config_paths

    screen_configs.append(screen_config)

def createCommand(arguments):
    command = arguments['run'].split(' ')
    command += screen_config['environments']
    command.append(arguments['python'])
    command.append(os.path.join(arguments['root'], 'utils', 'run.py'))

    for argument in ({'key' : 'config',            'prefix' : '--config'            },
                     {'key' : 'config_file',       'prefix' : '--config-file'       },
                     {'key' : 'config_path',       'prefix' : '--config-path'       },
                     {'key' : 'screen',            'prefix' : '--screen'            },
                     {'key' : 'blend_file',        'prefix' : '--blend-file'        },
                     ):

        if argument['key'] in arguments:
            value = arguments[argument['key']]
            if ' ' in value:
                value = '"' + value + '"'
            command.append(argument['prefix'] + "=" + value)

    if background:
        command.append('-b')

    if log_run:
        command.append('-l')

    return ' '.join(command)

commands = []
for screen_config in screen_configs:
    commands.append(createCommand(screen_config))

master   = commands[0]
commands = commands[1:]

delay = 0.5

def run(cmd):
    os.system(cmd)

def kill():
    if len(commands) > 0:
        for command in commands[::-1]:
            run(command)
        time.sleep(delay)
    run(master)

if state:
    if (previous_state):
        kill()
    master += ' -s'
    run(master)
    if len(commands) > 0:
        time.sleep(delay)
        for command in commands:
            run(command + ' -s')
else:
    kill()
